"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python class to create SPEXone Level-1 products

Copyright (c) 2019-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timedelta, timezone
from pathlib import Path, PurePosixPath

import numpy as np

from netCDF4 import Dataset

from .lib.attrs_def import attrs_def
from .lib.l1a_def import init_l1a
from .lib.l1b_def import init_l1b
from .lib.l1c_def import init_l1c

# - global parameters -------------------


# - class LV1io -------------------------
class Lv1io:
    """
    Generic class to create SPEXone Level-1 products

    Parameters
    ----------
    lv1_product: string
       Name of the Level-1 product
    append : boolean, optional
       Open file in append mode, parameter dims and inflight are ignored
       Default: False

    Attributes
    ----------
    product: pathlib.Path object
       Concrete path object to SPEXone Level-1 product
    inflight: boolean
       Flag to indicate data collected during in-flight of on-ground
    fid: netCDF5.Dataset object
       NetCDF4 Pointer to SPEXone Level-1 product
    ref_date: datetime.date object
       Reference date for 'seconds of day' parameters
    dset_stored: dict
       Number of items stored for all required netCDF4 variables

    Methods
    -------
    get_dims(ds_name)
       Returns size of dimension
    get_attr(attr_name, ds_name=None)
       Read data of an attribute, global or attached to a group or variable
    set_attr(attr_name, value, ds_name=None)
       Write data to an attribute, global or attached to a group or variable
    get_dset(ds_name)
       Read data of a netCDF4 variable
    set_dset(ds_name, value)
       Write data to a netCDF4 variable
    fill_global_attrs(level, orbit=-1, bin_size=None)

    Notes
    -----
    The engineering data should be extended, suggestions:
    * temperatures of a.o. detector, FEE, optica, obm, telescope
    * instrument settings: exposure time, dead time, frame time, coadding, ...
    """
    processing_level = 'unknown'
    dset_stored = {}

    def __init__(self, product: str, append=False, **kwargs):
        """
        Initialize access to a SPEXone Level-1 product

        Parameters
        ----------
        product : str
           name of the SPEXone Level-1 product
        append : bool
           do no clobber, but add new data to existing product
        """
        self.__epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)

        # initialize private class-attributes
        self.product = Path(product)
        self.inflight = True
        if 'inflight' in kwargs:
            self.inflight = kwargs['inflight']
        self.fid = None
        self.ref_date = None

        # initialize Level-1 product
        if not append:
            if self.processing_level == 'L1A':
                self.fid = init_l1a(product, **kwargs)
            elif self.processing_level == 'L1B':
                self.fid = init_l1b(product, **kwargs)
            elif self.processing_level == 'L1C':
                self.fid = init_l1c(product, **kwargs)
            else:
                raise KeyError('valid processing levels are: L1A, L1B or L1C')
        else:
            # open Level-1 product in append mode
            self.fid = Dataset(self.product, "r+")

            if 'reference_day' in self.fid.ncattrs():
                self.ref_date = datetime.fromisoformat(
                    self.fid.reference_day, tzinfo=timezone.utc)

            # store current length of the first dimension
            for key in self.dset_stored:
                self.dset_stored[key] = self.fid[key].shape[0]

    def __repr__(self):
        class_name = type(self).__name__
        return '{}({!r})'.format(class_name, self.product)

    def __iter__(self):
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self):
        """
        Close all resources (currently a placeholder function)
        """
        pass

    # ---------- PUBLIC FUNCTIONS ----------
    @property
    def epoch(self) -> datetime:
        """
        Provide epoch for SPEXone
        """
        return self.__epoch

    def get_dim(self, name: str):
        """
        Get size of a netCDF4 dimension
        """
        return self.fid.dimensions[name].size

    # ----- ATTRIBUTES --------------------
    def get_attr(self, name: str, ds_name=None):
        """
        Read data of an attribute, global or attached to a group or variable

        Parameters
        ----------
        name : string
           name of the attribute
        ds_name : string, optional
           name of dataset to which the attribute is attached

        Returns
        -------
        scalar or array_like
           value of attribute 'name', global or attached to dataset 'ds_name'
        """
        if ds_name is None:
            res = self.fid.getncattr(name)
        else:
            if ds_name not in self.fid.groups \
               and ds_name not in self.fid.variables:
                return None
            res = self.fid[ds_name].getncattr(name)

        if isinstance(res, bytes):
            return res.decode('ascii')

        return res

    def set_attr(self, name: str, value, ds_name=None) -> None:
        """
        Write data to an attribute, global or attached to a group or variable

        Parameters
        ----------
        name : string
           name of the attribute
        value : scalar, array_like
           value or values to be written
        ds_name : string
           name of group or dataset to which the attribute is attached
        """
        if ds_name is None:
            if isinstance(value, str):
                self.fid.setncattr(name, np.string_(value))
            else:
                self.fid.setncattr(name, value)
        else:
            grp_name = str(PurePosixPath(ds_name).parent)
            var_name = str(PurePosixPath(ds_name).name)
            if grp_name != '.':
                if var_name not in self.fid[grp_name].groups \
                   and var_name not in self.fid[grp_name].variables:
                    raise KeyError(
                        'ds_name {} not present in product'.format(ds_name))
            else:
                if var_name not in self.fid.groups \
                   and var_name not in self.fid.variables:
                    raise KeyError(
                        'ds_name {} not present in product'.format(ds_name))

            if isinstance(value, str):
                self.fid[ds_name].setncattr(name, np.string_(value))
            else:
                self.fid[ds_name].setncattr(name, value)

    # ----- VARIABLES --------------------
    def get_dset(self, name: str):
        """
        Read data of a netCDF4 variable

        Parameters
        ----------
        name : string
           name of dataset

        Returns
        -------
        scalar or array_like
           value of dataset 'name'
        """
        grp_name = str(PurePosixPath(name).parent)
        var_name = str(PurePosixPath(name).name)
        if grp_name != '.':
            if var_name not in self.fid[grp_name].variables:
                raise KeyError(
                    'dataset {} not present in Level-1 product'.format(name))
        else:
            if var_name not in self.fid.variables:
                raise KeyError(
                    'dataset {} not present in Level-1 product'.format(name))

        return self.fid[name][:]

    def set_dset(self, name: str, value, ibgn=-1) -> None:
        """
        Read data of a netCDF4 variable

        Parameters
        ----------
        name : string
           Name of Level-1 dataset
        value : scalar, array_like
           Value or values to be written
        ibgn : integer
           Index of the first (unlimited) dimension where to store the new data
           Default is to append the data
        """
        value = np.asarray(value)
        grp_name = str(PurePosixPath(name).parent)
        var_name = str(PurePosixPath(name).name)
        if grp_name != '.':
            if var_name not in self.fid[grp_name].variables:
                raise KeyError(
                    'dataset {} not present in Level-1 product'.format(name))
        else:
            if var_name not in self.fid.variables:
                raise KeyError(
                    'dataset {} not present in Level-1 product'.format(name))

        dims = self.fid[name].get_dims()
        if not dims:
            self.fid[name][...] = value
        elif dims[0].isunlimited():
            if ibgn < 0:
                ibgn = self.dset_stored[name]
            self.fid[name][ibgn:, ...] = value
        else:
            self.fid[name][...] = value
        self.dset_stored[name] += value.shape[0]

    # -------------------------
    def sec_of_day(self, utc_sec: int, frac_sec: float) -> float:
        """
        Parameters
        ----------
        utc_sec : numpy array
          seconds since 1970-01-01 (integer)
        frac_sec : numpy array
          fractional seconds (double)

        Returns
        -------
        - second of day
        """
        # get seconds since epoch
        utc0 = self.epoch + timedelta(seconds=int(utc_sec))

        # get seconds between midnight and epoch
        if self.ref_date is None:
            self.ref_date = datetime(year=utc0.year,
                                     month=utc0.month,
                                     day=utc0.day, tzinfo=timezone.utc)

        # return seconds since midnight
        return (utc0 - self.ref_date) / timedelta(seconds=1) + frac_sec

    # -------------------------
    def fill_global_attrs(self, orbit=-1, bin_size=None) -> None:
        """
        Define global attributes in the SPEXone Level-1 products

        Parameters
        ----------
        orbit_number: int
           Orbit revolution counter, default=-1
        bin_size: str, optional
           Size of the nadir footprint (cross-track), include unit: e.g. '5km'
        """
        dict_attrs = attrs_def(self.processing_level, self.inflight)
        dict_attrs['product_name'] = self.product.name
        dict_attrs['orbit_number'] = orbit
        if bin_size is not None:
            dict_attrs['bin_size_at_nadir'] = bin_size

        for key in dict_attrs:
            if dict_attrs[key] is not None:
                self.fid.setncattr(key, dict_attrs[key])


# - class L1Aio -------------------------
class L1Aio(Lv1io):
    """
    This class can be used to create a SPEXone Level-1A product

    Parameters
    ----------
    lv1_product: string
       Name of the Level-1A product
    append : boolean, optional
       Open file in append mode, parameter dims and inflight are ignored
       Default: False
    dims: dictionary, optional
       Provide size of various dimensions (L1A only).
       Default values:
            number_of_images : None     # number of image frames
            samples_per_image : 184000  # depends on binning table
            SC_records : None           # space-craft navigation records
            hk_packets : None           # number of HK tlm-packets
            wavelength : None
    inflight: boolean, optional
       In-flight data, only affects global attributes of L1A product.
       Default: False
    """
    processing_level = 'L1A'
    dset_stored = {
        '/science_data/detector_images': 0,
        '/science_data/detector_telemetry': 0,
        '/image_attributes/binning_table': 0,
        '/image_attributes/digital_offset': 0,
        '/image_attributes/nr_coadditions': 0,
        '/image_attributes/exposure_time': 0,
        '/image_attributes/image_CCSDS_sec': 0,
        '/image_attributes/image_CCSDS_usec': 0,
        '/image_attributes/image_time': 0,
        '/image_attributes/image_ID': 0,
        '/engineering_data/HK_telemetry': 0,
        '/engineering_data/temp_detector': 0,
        '/engineering_data/temp_optics': 0,
        '/engineering_data/HK_tlm_time': 0,
        '/navigation_data/adstate': 0,
        '/navigation_data/att_quat': 0,
        '/navigation_data/orb_pos': 0,
        '/navigation_data/orb_vel': 0,
        '/navigation_data/att_time': 0,
        '/navigation_data/orb_time': 0
    }

    def close(self):
        """
        close product and check if required datasets are filled with data
        """
        if self.fid is None:
            return

        if self.ref_date is not None \
           and 'reference_day' not in self.fid.ncattrs():
            self.fid.reference_day = self.ref_date.strftime('%Y-%m-%d')

        # check if atleast one dataset is updated
        if self.fid.dimensions['number_of_images'].size == 0:
            self.fid.close()
            self.fid = None
            return

        # check of all required dataset their sizes
        self.check_stored()

        # update coverage time
        # ToDo replace intg by master clock cycle or frame rate (FTI)
        intg = (self.fid['/image_attributes/exposure_time'][-1].data
                * self.fid['/image_attributes/nr_coadditions'][-1].data)

        img_sec = self.fid['/image_attributes/image_CCSDS_sec'][:].data
        img_usec = self.fid['/image_attributes/image_CCSDS_usec'][:].data

        time0 = (self.epoch
                 + timedelta(seconds=int(img_sec[0]))
                 + timedelta(microseconds=int(img_usec[0]))
                 - timedelta(seconds=intg))

        time1 = (self.epoch
                 + timedelta(seconds=int(img_sec[-1]))
                 + timedelta(microseconds=int(img_usec[-1])))

        self.fid.time_coverage_start = time0.isoformat(timespec='milliseconds')
        self.fid.time_coverage_end = time1.isoformat(timespec='milliseconds')

        self.fid.close()
        self.fid = None

    # -------------------------
    def check_stored(self):
        """
        Check variables with the same first dimension have equal sizes
        """
        warn_str = ('SPEX Level-1 format check [WARNING]:'
                    ' size of variable "{:s}" is wrong, only {:d} elements')

        # check image datasets
        dim_sz = self.get_dim('number_of_images')
        res = []
        key_list = [x for x in self.dset_stored
                    if (x.startswith('/science_data')
                        or x.startswith('/image_attributes'))]
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        indx = np.where(res != dim_sz)[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        # check house-keeping datasets
        dim_sz = self.get_dim('hk_packets')
        key_list = [x for x in self.dset_stored
                    if x.startswith('/engineering_data')]
        res = []
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        indx = np.where(res != dim_sz)[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        # check navigation datasets
        dim_sz = self.get_dim('SC_records')
        key_list = [x for x in self.dset_stored
                    if x.startswith('/navigation_data')]
        res = []
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        indx = np.where(res != dim_sz)[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

    # ---------- PUBLIC FUNCTIONS ----------
    def fill_mps(self, mps_data) -> None:
        """
        Write MPS information to L1A product

        Parameters
        ----------
        mps_data : numpy array
          Structured array with all MPS parameters

        Notes
        -----
        - mps_data as /science_data/detector_telemetry

        User friendly parameters:
        - exposure time as /image_attributes/exposure_time
        - nr_coadditions as /image_attributes/nr_coadditions
        """
        def digital_offset(mps) -> int:
            """
            Return digital offset including ADC offset
            """
            buff = mps['DET_OFFSET'].astype('i4')
            buff[buff >= 8192] -= 16384

            return buff + 70

        def exposure_time(mps, mcp=1e-7) -> float:
            """
            Return detector exposure_time
            """
            exptime = 129 * (0.43 * mps['DET_FOTLEN'] + mps['DET_EXPTIME'])

            return mcp * exptime

        self.set_dset('/science_data/detector_telemetry', mps_data)

        self.set_dset('/image_attributes/binning_table',
                      mps_data['REG_BINNING_TABLE'])
        self.set_dset('/image_attributes/digital_offset',
                      digital_offset(mps_data))
        self.set_dset('/image_attributes/exposure_time',
                      exposure_time(mps_data))
        self.set_dset('/image_attributes/nr_coadditions',
                      mps_data['REG_NCOADDFRAMES'])

    def fill_time(self, sec_of_day, reference_day, *, leap_seconds=0) -> None:
        """
        Write TM time information to L1A product

        Parameters
        ----------
        sec_of_day : numpy array (float)
           seconds since midnight
        reference_day : datetime.datetime
           midnight
        leap_second : integer
          leap seconds since 1970, use only when input are TAI seconds
        """
        self.set_dset('/image_attributes/image_time', sec_of_day)

        utc_sec = sec_of_day.astype('u4') + leap_seconds \
            + (reference_day - self.epoch).total_seconds()
        frac_sec = np.round(1e6 * (sec_of_day % 1)).astype('i4')

        self.set_dset('/image_attributes/image_CCSDS_sec', utc_sec)
        self.set_dset('/image_attributes/image_CCSDS_usec', frac_sec)

    def fill_gse(self, reference=None) -> None:
        """
        Write EGSE/OGSE data to L1A product

        Parameters
        ----------
        reference : dict, optional
           biweight value and spread of the signal measured during the
           measurement by a reference detector.
           Expected dictionary keys: 'value', 'error'
        """
        if reference is not None:
            dset = self.fid.createVariable('/gse_data/reference_signal',
                                           'f8', ())
            dset.long_name = "biweight median of reference-detector signal"
            dset.comment = "t_sat = min([2.28e-9 / S_reference, 30])"
            dset.units = 'A'
            dset[:] = reference['value']

            dset = self.fid.createVariable('/gse_data/reference_error',
                                           'f8', ())
            dset.long_name = "biweight spread of reference-detector signal"
            dset.units = 'A'
            dset[:] = reference['error']


# - class L1Bio -------------------------
class L1Bio(Lv1io):
    """
    This class can be used to create a SPEXone Level-1B product

    Parameters
    ----------
    lv1_product: string
       Name of the Level-1B product
    append : boolean, optional
       Open file in append mode, parameter dims and inflight are ignored
       Default: False
    number_of_images: int
       Number of images used as input to generate the L1B product.
       Default is None, then this dimension is UNLIMITED.
    spatial_samples: int
       Total number of spatial samples from all viewports, default is 200
    """
    processing_level = 'L1B'
    dset_stored = {
        '/SENSOR_VIEWS_BANDS/viewport_index': 0,
        '/SENSOR_VIEWS_BANDS/view_angles': 0,
        '/SENSOR_VIEWS_BANDS/intensity_wavelengths': 0,
        '/SENSOR_VIEWS_BANDS/intensity_bandpasses': 0,
        '/SENSOR_VIEWS_BANDS/polarization_wavelengths': 0,
        '/SENSOR_VIEWS_BANDS/polarization_bandpasses': 0,
        '/SENSOR_VIEWS_BANDS/intensity_f0': 0,
        '/SENSOR_VIEWS_BANDS/polarization_f0': 0,
        '/BIN_ATTRIBUTES/image_time': 0,
        '/GEOLOCATION_DATA/latitude': 0,
        '/GEOLOCATION_DATA/longitude': 0,
        '/GEOLOCATION_DATA/altitude': 0,
        '/GEOLOCATION_DATA/altitude_variability': 0,
        '/GEOLOCATION_DATA/sensor_azimuth': 0,
        '/GEOLOCATION_DATA/sensor_zenith': 0,
        '/GEOLOCATION_DATA/solar_azimuth': 0,
        '/GEOLOCATION_DATA/solar_zenith': 0,
        '/OBSERVATION_DATA/I': 0,
        '/OBSERVATION_DATA/I_noise': 0,
        '/OBSERVATION_DATA/q': 0,
        '/OBSERVATION_DATA/q_noise': 0,
        '/OBSERVATION_DATA/u': 0,
        '/OBSERVATION_DATA/u_noise': 0,
        '/OBSERVATION_DATA/AoLP': 0,
        '/OBSERVATION_DATA/AoLP_noise': 0,
        '/OBSERVATION_DATA/DoLP': 0,
        '/OBSERVATION_DATA/DoLP_noise': 0
    }

    def close(self):
        """
        close product and check if required datasets are filled with data
        """
        if self.fid is None:
            return

        if self.ref_date is not None \
           and 'reference_day' not in self.fid.ncattrs():
            self.fid.reference_day = self.ref_date.strftime('%Y-%m-%d')

        # check if atleast one dataset is updated
        if self.fid.dimensions['bins_along_track'].size == 0:
            self.fid.close()
            self.fid = None
            return

        # check of all required dataset their sizes
        self.check_stored()

        # update coverage time
        # ToDo epoch should be reference_day
        secnd = self.fid['/BIN_ATTRIBUTES/image_time'][0].data
        time0 = (self.epoch
                 + timedelta(seconds=int(secnd))
                 + timedelta(microseconds=int(secnd % 1)))

        secnd = self.fid['/BIN_ATTRIBUTES/image_time'][-1].data
        time1 = (self.epoch
                 + timedelta(seconds=int(secnd))
                 + timedelta(microseconds=int(secnd % 1)))

        self.fid.time_coverage_start = time0.isoformat(timespec='milliseconds')
        self.fid.time_coverage_end = time1.isoformat(timespec='milliseconds')

        self.fid.close()
        self.fid = None

    # -------------------------
    def check_stored(self):
        """
        Check variables with the same first dimension have equal sizes
        """
        warn_str = ('SPEX Level-1 format check [WARNING]:'
                    ' size of variable "{:s}" is wrong, only {:d} elements')

        # check datasets in group /SENSOR_VIEWS_BANDS
        dim_sz = self.get_dim('number_of_views')
        res = []
        key_list = [x for x in self.dset_stored
                    if x.startswith('/SENSOR_VIEWS_BANDS')]
        for key in key_list:
            if key == '/SENSOR_VIEWS_BANDS/viewport_index':
                continue
            res.append(self.dset_stored[key])
        res = np.array(res)
        indx = np.where(res != dim_sz)[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        # check datasets in all other groups
        dim_sz = self.get_dim('bins_along_track')
        res = []
        key_list = [x for x in self.dset_stored
                    if not x.startswith('/SENSOR_VIEWS_BANDS')]
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        indx = np.where(res != dim_sz)[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        for ii, key in enumerate(self.dset_stored):
            print(ii, key, self.dset_stored[key])

    # ---------- PUBLIC FUNCTIONS ----------


# - class L1Cio -------------------------
class L1Cio(Lv1io):
    """
    This class can be used to create a SPEXone Level-1C product

    Parameters
    ----------
    lv1_product: string
       Name of the Level-1C product
    append : boolean, optional
       Open file in append mode, parameter dims and inflight are ignored
       Default: False
    number_of_images: int
       Number of images used as input to generate the L1B product.
       Default is None, then this dimension is UNLIMITED.
    """
    processing_level = 'L1C'
    dset_stored = {
        '/SENSOR_VIEWS_BANDS/view_angles': 0,
        '/SENSOR_VIEWS_BANDS/intensity_wavelengths': 0,
        '/SENSOR_VIEWS_BANDS/intensity_bandpasses': 0,
        '/SENSOR_VIEWS_BANDS/polarization_wavelengths': 0,
        '/SENSOR_VIEWS_BANDS/polarization_bandpasses': 0,
        '/SENSOR_VIEWS_BANDS/intensity_f0': 0,
        '/SENSOR_VIEWS_BANDS/polarization_f0': 0,
        '/BIN_ATTRIBUTES/nadir_view_time': 0,
        '/BIN_ATTRIBUTES/view_time_offsets': 0,
        '/GEOLOCATION_DATA/latitude': 0,
        '/GEOLOCATION_DATA/longitude': 0,
        '/GEOLOCATION_DATA/altitude': 0,
        '/GEOLOCATION_DATA/altitude_variability': 0,
        '/GEOLOCATION_DATA/sensor_azimuth': 0,
        '/GEOLOCATION_DATA/sensor_zenith': 0,
        '/GEOLOCATION_DATA/solar_azimuth': 0,
        '/GEOLOCATION_DATA/solar_zenith': 0,
        '/OBSERVATION_DATA/obs_per_view': 0,
        '/OBSERVATION_DATA/I': 0,
        '/OBSERVATION_DATA/I_noise': 0,
        '/OBSERVATION_DATA/I_polsample': 0,
        '/OBSERVATION_DATA/I_polsample_noise': 0,
        '/OBSERVATION_DATA/Q': 0,
        '/OBSERVATION_DATA/Q_noise': 0,
        '/OBSERVATION_DATA/U': 0,
        '/OBSERVATION_DATA/U_noise': 0,
        '/OBSERVATION_DATA/q': 0,
        '/OBSERVATION_DATA/q_noise': 0,
        '/OBSERVATION_DATA/u': 0,
        '/OBSERVATION_DATA/u_noise': 0,
        '/OBSERVATION_DATA/DoLP': 0,
        '/OBSERVATION_DATA/DoLP_noise': 0
    }

    def close(self):
        """
        close product and check if required datasets are filled with data
        """
        if self.fid is None:
            return

        if self.ref_date is not None \
           and 'reference_day' not in self.fid.ncattrs():
            self.fid.reference_day = self.ref_date.strftime('%Y-%m-%d')

        # check if atleast one dataset is updated
        if self.fid.dimensions['bins_along_track'].size == 0:
            self.fid.close()
            self.fid = None
            return

        # check of all required dataset their sizes
        self.check_stored()

        # update coverage time
        # ToDo epoch should be reference_day
        secnd = self.fid['/BIN_ATTRIBUTES/nadir_view_time'][0].data
        time0 = (self.epoch
                 + timedelta(seconds=int(secnd))
                 + timedelta(microseconds=int(secnd % 1)))

        secnd = self.fid['/BIN_ATTRIBUTES/nadir_view_time'][-1].data
        time1 = (self.epoch
                 + timedelta(seconds=int(secnd))
                 + timedelta(microseconds=int(secnd % 1)))

        self.fid.time_coverage_start = time0.isoformat(timespec='milliseconds')
        self.fid.time_coverage_end = time1.isoformat(timespec='milliseconds')

        self.fid.close()
        self.fid = None

    # -------------------------
    def check_stored(self):
        """
        Check variables with the same first dimension have equal sizes
        """
        warn_str = ('SPEX Level-1 format check [WARNING]:'
                    ' size of variable "{:s}" is wrong, only {:d} elements')

        # check datasets in group /SENSOR_VIEWS_BANDS
        dim_sz = self.get_dim('number_of_views')
        res = []
        key_list = [x for x in self.dset_stored
                    if x.startswith('/SENSOR_VIEWS_BANDS')]
        for key in key_list:
            if key == '/SENSOR_VIEWS_BANDS/viewport_index':
                continue
            res.append(self.dset_stored[key])
        res = np.array(res)
        indx = np.where(res != dim_sz)[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        # check datasets in all other groups
        dim_sz = self.get_dim('bins_along_track')
        res = []
        key_list = [x for x in self.dset_stored
                    if not x.startswith('/SENSOR_VIEWS_BANDS')]
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        indx = np.where(res != dim_sz)[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        for ii, key in enumerate(self.dset_stored):
            print(ii, key, self.dset_stored[key])

    # ---------- PUBLIC FUNCTIONS ----------



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Create a valid PACE SPEXone Level-1C product to estimate the size of a
science orbit product

Copyright (c) 2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timezone

import numpy as np

from pyspex.lv1_io import L1Cio

# - global parameters ------------------------------


# - local functions --------------------------------

# - main function ----------------------------------
def main():
    """
    Generate a SPEXone inflight Level-1C product
    """
    start_time = datetime(2023, 1, 15, 12, 34, 56, tzinfo=timezone.utc)
    midnight = start_time.replace(hour=0, minute=0, second=0, microsecond=0)
    duration = 2880  # seconds

    # define product: name, level, version, start of time-coverage
    prod_level = 'L1C'
    bin_size = '5km'
    prod_version = 'V1.0'
    l1c_flname = 'PACE_SPEX.{:s}.{:s}.{:s}.{:s}.nc'.format(
        start_time.strftime('%Y%m%dT%H%M%S'),
        prod_level, bin_size, prod_version)

    with L1Cio(l1c_flname, dims={}) as l1c:
        secnds = (1 + np.arange(3 * duration, dtype=float)) / 3\
            + (start_time - midnight).total_seconds()

        # get dimension sizes
        n_view = l1c.get_dim('number_of_views')
        n_intens = l1c.get_dim('intensity_bands_per_view')
        n_polar = l1c.get_dim('polarization_bands_per_view')
        n_across = l1c.get_dim('bins_across_track')

        # group SENSOR_VIEWS_BANDS
        l1c.set_dset('/SENSOR_VIEWS_BANDS/view_angles',
                     np.array([-58., -30., 0., 30., 58.]))
        l1c.set_dset('/SENSOR_VIEWS_BANDS/intensity_wavelengths',
                     np.arange(n_view * n_intens).reshape(n_view, n_intens))
        l1c.set_dset('/SENSOR_VIEWS_BANDS/intensity_bandpasses',
                     np.arange(n_view * n_intens).reshape(n_view, n_intens))
        l1c.set_dset('/SENSOR_VIEWS_BANDS/polarization_wavelengths',
                     np.arange(n_view * n_polar).reshape(n_view, n_polar))
        l1c.set_dset('/SENSOR_VIEWS_BANDS/polarization_bandpasses',
                     np.arange(n_view * n_polar).reshape(n_view, n_polar))
        l1c.set_dset('/SENSOR_VIEWS_BANDS/intensity_f0',
                     np.arange(n_view * n_intens).reshape(n_view, n_intens))
        l1c.set_dset('/SENSOR_VIEWS_BANDS/polarization_f0',
                     np.arange(n_view * n_polar).reshape(n_view, n_polar))

        for ii in range(3 * duration):
            # group BIN_ATTRIBUTES
            l1c.set_dset('/BIN_ATTRIBUTES/nadir_view_time', [secnds[ii]])
            data = np.zeros((1, n_across), dtype=float) \
                - ii / (6 * duration)
            l1c.set_dset('/BIN_ATTRIBUTES/view_time_offsets', data)

            # group GEOLOCATION_DATA
            data = np.zeros((1, n_across), dtype=float) \
                + ii / duration
            l1c.set_dset('/GEOLOCATION_DATA/latitude', data)
            l1c.set_dset('/GEOLOCATION_DATA/longitude', data)
            l1c.set_dset('/GEOLOCATION_DATA/altitude', data)
            l1c.set_dset('/GEOLOCATION_DATA/altitude_variability', data)
            l1c.set_dset('/GEOLOCATION_DATA/solar_azimuth', data)
            l1c.set_dset('/GEOLOCATION_DATA/solar_zenith', data)
            data = np.zeros((1, n_across, n_view), dtype=float) \
                - ii / (6 * duration)
            l1c.set_dset('/GEOLOCATION_DATA/sensor_azimuth', data)
            l1c.set_dset('/GEOLOCATION_DATA/sensor_zenith', data)

            # group OBSERVATION_DATA
            data = np.ones((1, n_across, n_view), dtype=float) * 4
            l1c.set_dset('/OBSERVATION_DATA/obs_per_view', data)
            data = np.ones((1, n_across, n_view, n_intens), dtype=float) \
                + ii / duration
            l1c.set_dset('/OBSERVATION_DATA/I', data)
            l1c.set_dset('/OBSERVATION_DATA/I_noise', data)
            data = np.ones((1, n_across, n_view, n_polar), dtype=float) \
                + ii / duration
            l1c.set_dset('/OBSERVATION_DATA/I_polsample', data)
            l1c.set_dset('/OBSERVATION_DATA/I_polsample_noise', data)
            l1c.set_dset('/OBSERVATION_DATA/Q', data)
            l1c.set_dset('/OBSERVATION_DATA/Q_noise', data)
            l1c.set_dset('/OBSERVATION_DATA/U', data)
            l1c.set_dset('/OBSERVATION_DATA/U_noise', data)
            l1c.set_dset('/OBSERVATION_DATA/q', data)
            l1c.set_dset('/OBSERVATION_DATA/q_noise', data)
            l1c.set_dset('/OBSERVATION_DATA/u', data)
            l1c.set_dset('/OBSERVATION_DATA/u_noise', data)
            l1c.set_dset('/OBSERVATION_DATA/DoLP', data)
            l1c.set_dset('/OBSERVATION_DATA/DoLP_noise', data)

        # Global attributes
        l1c.fill_global_attrs(orbit=12345, bin_size=bin_size)


# --------------------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pySpexCal

https://gitlab.sron.nl/Richardh/pySpexCal

Generate a L1A product with simulated data
- one full science orbit (50% van 5900 sec), max 24 Gbit
- [SPX1_FLIGHT_ideal_202003] 4 scenes (‘ocean_cloud’, ‘ocean_moderate_aot’,
 ‘soil_moderate_aot’ en ‘vegetation_moderate_aot’) at 3 solar zenith angles
 (10, 40 en 70 graden)
- sampling at 3 Hz (at 50 ms, 5x coadding + 16.6 dead-time)
- binned at 2x2 with lineskipped (1024x645)

Implementation:
1. read simulated datasets (original format, lineskip)
2. generate 3x 5900 // 2 images SZA 70, 40, 10, 10, 40, 70
   each SZA filled with all 4 scenes

   >>> rng = np.random.default_rng()
   >>> indx = np.arange(3 * 5900 // 10)
   >>> rng.shuffle(indx)
   >>> indx.repeat(5)

3. fill groups: /image_attributes, /science_data of L1A product
4. fill groups: /navigation_data, /engineering_data
5. leave group: /gse_data empty

Copyright (c) 2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse
from datetime import datetime, timezone
from pathlib import Path

import h5py
import numpy as np

from pyspex.lv1_io import L1Aio

FLNAME_OUT = 'SPX1_TEST_L1A_simulated_orbit.nc'

# - global parameters ------------------------------
MIDNIGHT = datetime(2020, 7, 4, tzinfo=timezone.utc)


# - local functions --------------------------------


# --------------------------------------------------
def initialize_l1a_product(l1a_nav_product: str, bin_tbl: int) -> None:
    """
    Initialize an L1A product with navigation data

    Parameters
    ----------
    l1a_nav_product : string
    bin_tbl : integer
    """
    ckd_dir = '/nfs/SPEXone/share/ckd'
    if not Path(ckd_dir).is_dir():
        ckd_dir = '/data/richardh/SPEXone/share/ckd'
    # only read the latest version of the binning-table CKD
    bin_tbl_fl = sorted(list(Path(ckd_dir).glob('SPX1_OCAL_L1A_TBL_*.nc')))
    if not bin_tbl_fl:
        raise FileNotFoundError('No binning table found')

    # read binning table information
    with h5py.File(bin_tbl_fl[-1], 'r') as fid:
        gid = fid['Table_{:02d}'.format(bin_tbl)]
        # enabled_lines = gid.attrs['enabled_lines']
        n_samples = gid['binning_table'].attrs['valid_max'] + 1

    # read navigation file information
    with h5py.File(l1a_nav_product, 'r') as fid:
        # SC_hkt_block = fid['SC_hkt_block'].size
        nav_records = fid['SC_records'].size
        # quaternion_elements = fid['quaternion_elements'].size
        # vector_elements = fid['vector_elements'].size
        gid = fid['navigation_data']
        att_time = gid['att_time'][:]
        att_quat = gid['att_quat'][:]
        orb_time = gid['orb_time'][:]
        orb_pos = gid['orb_pos'][:]
        if np.mean(np.abs(orb_pos)) < 5000:
            orb_pos *= 1000              # convert km -> meters
        orb_vel = gid['orb_vel'][:]
        if np.mean(np.linalg.norm(orb_vel, axis=1)) < 8:
            orb_vel *= 1000              # convert km -> meters
        adstate = gid['adstate'][:]

    # create L1A product
    # ToDo fix the output product name
    dims = {'number_of_images': None,
            'samples_per_image': n_samples,
            'SC_records': nav_records,
            'hk_packets': None}

    with L1Aio(FLNAME_OUT, dims=dims, inflight=True) as l1a:
        # navigation data
        l1a.set_dset('/navigation_data/att_time', att_time)
        l1a.set_dset('/navigation_data/att_quat', att_quat)
        l1a.set_dset('/navigation_data/orb_time', orb_time)
        l1a.set_dset('/navigation_data/orb_pos', orb_pos)
        l1a.set_dset('/navigation_data/orb_vel', orb_vel)
        l1a.set_dset('/navigation_data/adstate', adstate)

        # Global attributes
        l1a.fill_global_attrs(orbit=12345, bin_size='2.5km')
        l1a.set_attr('history', 'simulated orbit file')

    return min(att_time[-1], orb_time[-1]) - max(att_time[0], orb_time[0])


def add_measurements(l1a_prod_list, repeats: int, sampling=3) -> None:
    """
    Add simulated measurements to L1A product

    Parameters
    ----------
    l1a_prod_list : list of strings
    repeats : integer
    sampling : integer
    """
    # read data
    coad = []
    texp = []
    images = []
    temp_det = []
    temp_opt = []
    for flname in l1a_prod_list:
        with h5py.File(flname, 'r') as fid:
            mps_dtype = fid['/science_data/detector_telemetry'].dtype
            coad.append(fid['/image_attributes/nr_coadditions'][:])
            texp.append(fid['/image_attributes/exposure_time'][:])
            images.append(fid['/science_data/detector_images'][:])
            temp_det.append(fid['/engineering_data/temp_detector'][:])
            temp_opt.append(fid['/engineering_data/temp_optics'][:])

    coad = np.stack(coad)
    texp = np.stack(texp)
    images = np.stack(images)
    temp_det = np.stack(temp_det)
    temp_opt = np.stack(temp_opt)

    # create random index for sciences
    rng = np.random.default_rng()
    arr = np.arange(sampling * repeats / coad.shape[1], dtype=int)
    rng.shuffle(arr)
    indx = arr % len(coad)

    # repeat data to fill measurement block
    coad = coad[indx, :].reshape(-1)
    texp = texp[indx, :].reshape(-1)
    samples_per_image = images.shape[-1]
    images = images[indx, :, :].reshape(-1, samples_per_image)
    n_images = images.shape[0]
    temp_det = temp_det[indx[::sampling], :].reshape(-1)
    temp_opt = temp_opt[indx[::sampling], :].reshape(-1)

    # generate telemetry data
    telemetry = np.zeros(n_images, dtype=mps_dtype)

    with L1Aio(FLNAME_OUT, append=True) as l1a:
        att_time = int(l1a.get_dset('/navigation_data/att_time')[0])

        # define timing of image data
        ibgn = l1a.get_dim('number_of_images')
        print('ibgn (images): ', ibgn)
        img_time = att_time + (ibgn + np.arange(n_images)) / sampling

        # write datasets in /science_data
        l1a.set_dset('/science_data/detector_images', images)
        l1a.set_dset('/science_data/detector_telemetry', telemetry)

        # write datasets in /image_attributes
        l1a.set_dset('/image_attributes/exposure_time', texp)
        l1a.set_dset('/image_attributes/nr_coadditions', coad)
        l1a.set_dset('/image_attributes/image_ID',
                     np.arange(n_images) + ibgn)
        l1a.set_dset('/image_attributes/digital_offset',
                     np.zeros(n_images))
        l1a.set_dset('/image_attributes/binning_table',
                     np.full(n_images, 3))
        l1a.fill_time(img_time, MIDNIGHT)

        # define timing of house-keeping data
        ibgn = l1a.get_dim('hk_packets')
        print('ibgn (engineering): ', ibgn)
        hk_time = att_time + (ibgn + np.arange(temp_det.size))

        # write datasets in /engineering_data
        l1a.set_dset('/engineering_data/HK_tlm_time', hk_time)
        l1a.set_dset('/engineering_data/temp_detector', temp_det)
        l1a.set_dset('/engineering_data/temp_optics', temp_opt)


# - main functions --------------------------------
def main():
    """
    main function
    """
    # parse command-line parameters
    parser = argparse.ArgumentParser(
        description='generate SPEXone L1A product simulating a science orbit')
    parser.add_argument('--navigation_data', default=None,
                        help=('navigation data used to initialize the L1A file,'
                              ' provide path to an existing L1A product'))
    parser.add_argument('--binTableID', default=3, type=int,
                        help=('provide binning table ID, default=3'))
    parser.add_argument('--repeats', default=1, type=int,
                        help=('specify number of repeats of each measurement,'
                              ' default=1'))
    parser.add_argument('--measurement_data', default=None, nargs='*',
                        help=('provide path one or more L1A products with,'
                              ' measurement data used to fill the L1A product'))
    parser.add_argument('--verbose', action='store_true', default=False)
    args = parser.parse_args()
    if args.verbose:
        print(args)

    if args.navigation_data is not None:
        duration = initialize_l1a_product(args.navigation_data, args.binTableID)
        print('Coverage of the navigation data is {} seconds'.format(duration))

    if args.measurement_data is not None:
        add_measurements(args.measurement_data, args.repeats, sampling=3)


# --------------------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation to convert SPEXone DEM measurements to L1A format

Copyright (c) 2019-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse

from datetime import datetime, timedelta, timezone
from pathlib import Path

import numpy as np

from pyspex import spx_product
from pyspex.ccsds_io import CCSDSio
from pyspex.lv1_io import L1Aio

# - global parameters ------------------------------
LAUNCH_DATE = datetime(2022, 11, 2, tzinfo=timezone.utc)
EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)
LEAP_SECONDS = 27                          # ToDo use package leapseconds.py

# - local functions --------------------------------


# - main function ----------------------------------
def main():
    """
    main function
    """
    # parse command-line parameters
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description='create SPEXone Level-1A product from CCSDS packages (L0)')
    parser.add_argument('--tmtc_issue', default=12, type=int, help=(
        'Specify issue of the TMTC handbook, default=12\n'
        ' < 12: Science data header format before 15-May-2020\n'
        ' 12: append 6 bytes timestamp after MPS data [15-May-2020]'))
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('file_list', nargs='+',
                        help=("provide names of one or more files with"
                              " CCSDS packages of the same measurement"))
    args = parser.parse_args()
    if args.verbose:
        print(args)

    packet_list = ()
    ccsds = CCSDSio(tmtc_issue=args.tmtc_issue, verbose=args.verbose)
    for flname in sorted(args.file_list):
        if not Path(flname).is_file():
            raise FileNotFoundError(
                'File {} does not exist'.format(flname))
        if Path(flname).suffix == '.H':
            print('[INFO]: skip CCSDS header file')
            continue

        packet_list += ccsds.read(flname)

    if not packet_list:
        print('[WARNING]: no CCSDS packets found')
        return

    # combine segmented packages
    packets = ccsds.group(packet_list)

    tstamp = []
    mps_data = []
    image_id = []
    images = []
    for packet in packets:
        utc_sec = packet['secondary_header']['tai_sec'] - LEAP_SECONDS
        frac_sec = packet['secondary_header']['sub_sec'] / 2**16
        tstamp.append(EPOCH + timedelta(seconds=utc_sec + frac_sec))

        image_id.append(packet['primary_header']['sequence'] & 0x3fff)
        mps_data.append(packet['mps'])
        images.append(packet['image_data'])

    image_id = np.array(image_id)
    mps_data = np.array(mps_data)
    images = np.array(images)
    if args.verbose:
        print(images.size
              if images.ndim == 1 else images.shape[1], images.shape)

    # generate name of L1A product
    if tstamp[0] < LAUNCH_DATE:
        prod_name = spx_product.prod_name(
            tstamp[0], msm_id=Path(args.file_list[0]).stem)
        inflight = False
    else:
        prod_name = spx_product.prod_name(tstamp[0])
        inflight = True

    # pylint: disable=unsubscriptable-object
    n_frame = 1 if images.ndim == 1 else images.shape[0]
    n_sample = images.size if images.ndim == 1 else images.shape[1]
    dims = {'number_of_images': n_frame,
            'samples_per_image': n_sample,
            'SC_records': None,
            'tlm_packets': None}

    # convert timestamps to seconds per day
    secnds = np.empty(len(tstamp), dtype=float)
    midnight = tstamp[0].replace(hour=0, minute=0, second=0, microsecond=0)
    for ii, tval in enumerate(tstamp):
        secnd = (tval - midnight).total_seconds()
        # time-of-measurement provided at end of integration time
        # secnd -= texp[ii] * mps_data[ii]['REG_NCOADDFRAMES']
        secnds[ii] = secnd

    #
    # Generate L1A product
    #   ToDo: correct value for measurement_type & viewport
    with L1Aio(prod_name, dims=dims, inflight=inflight) as l1a:
        l1a.set_dset('/science_data/detector_images', images)
        l1a.fill_mps(mps_data)

        l1a.fill_time(secnds, midnight)
        l1a.set_dset('/image_attributes/image_ID', image_id)

        l1a.set_dset('/engineering_data/HK_tlm_time', secnds)

        # Global attributes
        l1a.fill_global_attrs()
        l1a.set_attr('input_files',
                     [Path(x).name for x in args.file_list])


# --------------------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation of the PACE SPEX Level-1C product

Copyright (c) 2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from netCDF4 import Dataset

# - global parameters ------------------------------


# - local functions --------------------------------


# - main function ----------------------------------
def init_l1c(l1c_flname: str, dims: dict):
    """
    Create an empty PACE SPEX Level-1C product

    Parameters
    ----------
    l1c_flname : string
       Name of Level-1C product
    dims :   dictionary
       Provide length of the Level-1B dimensions
       Default values:
            number_of_images : None     # number of image frames
    """
    # check function parameters
    if not isinstance(dims, dict):
        raise TypeError("dims should be a dictionary")

    # initialize dimensions
    number_of_images = None
    n_views = 5
    n_bins_intens = 400
    n_bins_polar = 50
    n_bins_across = 40

    if 'number_of_images' in dims:
        number_of_images = dims['number_of_images']

    # create/overwrite netCDF4 product
    rootgrp = Dataset(l1c_flname, "w")

    # create global dimensions
    _ = rootgrp.createDimension('number_of_views', n_views)
    _ = rootgrp.createDimension('intensity_bands_per_view', n_bins_intens)
    _ = rootgrp.createDimension('polarization_bands_per_view', n_bins_polar)
    _ = rootgrp.createDimension('bins_across_track', n_bins_across)
    _ = rootgrp.createDimension('bins_along_track', number_of_images)

    # create groups and all variables with attributes
    sgrp = rootgrp.createGroup('SENSOR_VIEWS_BANDS')
    dset = sgrp.createVariable('view_angles', 'f4', ('number_of_views',))
    dset.long_name = 'along track view zenith angles at sensor'
    dset.units = 'degrees'
    dset.comment = ('view_angles is defined at the sensor, as it provides'
                    ' a swath independent value at TOA.')
    dset = sgrp.createVariable('intensity_wavelengths', 'f4',
                               ('number_of_views', 'intensity_bands_per_view'))
    dset.long_name = 'wavelength at center of intensity bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('intensity_bandpasses', 'f4',
                               ('number_of_views', 'intensity_bands_per_view'))
    dset.long_name = 'FWHM of intensity bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('polarization_wavelengths', 'f4',
                               ('number_of_views',
                                'polarization_bands_per_view'))
    dset.long_name = 'wavelength at center of polarization bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('polarization_bandpasses', 'f4',
                               ('number_of_views',
                                'polarization_bands_per_view'))
    dset.long_name = 'FWHM of polarization bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('intensity_f0', 'f4',
                               ('number_of_views', 'intensity_bands_per_view'))
    dset.long_name = 'Solar irradiance on intensity wavelength grid'
    dset.units = 'W.m-2'
    dset.comment = ('Spectral response function convolved mean solar flux'
                    ' at each intensity band and view.')
    dset = sgrp.createVariable('polarization_f0', 'f4',
                               ('number_of_views',
                                'polarization_bands_per_view'))
    dset.long_name = 'Solar irradiance on polarization wavelength grid'
    dset.units = 'W.m-2'
    dset.comment = ('Spectral response function convolved mean solar flux'
                    ' at each polarization band and view.')

    sgrp = rootgrp.createGroup('BIN_ATTRIBUTES')
    chunksizes = None if number_of_images is not None else (512,)
    dset = sgrp.createVariable('nadir_view_time', 'f8', ('bins_along_track',),
                               chunksizes=chunksizes)
    dset.long_name = 'nadir time (seconds of day)'
    dset.valid_min = 0
    dset.valid_max = 86400.999999
    dset.units = 'seconds'
    dset.reference = 'yyyy-mm-ddT00:00:00'
    chunksizes = None if number_of_images is not None else (512, n_bins_across)
    dset = sgrp.createVariable('view_time_offsets', 'f8',
                               ('bins_along_track', 'bins_across_track'),
                               chunksizes=chunksizes)
    dset.long_name = 'offset to nadir view time'
    dset.units = 'seconds'

    sgrp = rootgrp.createGroup('GEOLOCATION_DATA')
    dset = sgrp.createVariable('latitude', 'f4',
                               ('bins_along_track', 'bins_across_track'),
                               chunksizes=chunksizes)
    dset.standard_name = 'latitude'
    dset.long_name = 'latitude'
    dset.valid_min = -90
    dset.valid_max = 90
    dset.units = 'degrees_north'
    dset = sgrp.createVariable('longitude', 'f4',
                               ('bins_along_track', 'bins_across_track'),
                               chunksizes=chunksizes)
    dset.standard_name = 'longitude'
    dset.long_name = 'longitude'
    dset.valid_min = -180
    dset.valid_max = 180
    dset.units = 'degrees_east'
    dset = sgrp.createVariable('altitude', 'f4',
                               ('bins_along_track', 'bins_across_track'),
                               chunksizes=chunksizes)
    dset.standard_name = 'altitude'
    dset.long_name = "height above mean sea level"
    dset.units = "meters"
    dset.positive = "up"
    dset.axis = "Z"
    dset = sgrp.createVariable('altitude_variability', 'f4',
                               ('bins_along_track', 'bins_across_track'),
                               chunksizes=chunksizes)
    dset.long_name = 'altitude (stdev)'
    dset.units = 'meters'
    dset = sgrp.createVariable('solar_azimuth', 'f4',
                               ('bins_along_track', 'bins_across_track'),
                               chunksizes=chunksizes)
    dset.long_name = 'solar azimuth angle'
    dset.units = 'degrees'
    dset = sgrp.createVariable('solar_zenith', 'f4',
                               ('bins_along_track', 'bins_across_track'),
                               chunksizes=chunksizes)
    dset.long_name = 'solar zenith angle'
    dset.units = 'degrees'

    chunksizes = None if number_of_images is not None else \
        (512, n_bins_across, n_views)
    dset = sgrp.createVariable('sensor_azimuth', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views'),
                               chunksizes=chunksizes)
    dset.long_name = 'sensor azimuth angle'
    dset.units = 'degrees'
    dset = sgrp.createVariable('sensor_zenith', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views'),
                               chunksizes=chunksizes)
    dset.long_name = 'sensor zenith angle'
    dset.units = 'degrees'

    chunksizes = None if number_of_images is not None else (512, 40, 5)
    sgrp = rootgrp.createGroup('OBSERVATION_DATA')
    dset = sgrp.createVariable('obs_per_view', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views'),
                               chunksizes=chunksizes)
    dset.long_name = 'observations per view'
    dset.units = '1'
    dset.comment = "Observations contributing to bin from each view"

    chunksizes = None if number_of_images is not None else \
        (32, n_bins_across, n_views, n_bins_intens)
    dset = sgrp.createVariable('I', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views', 'intensity_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'I Stokes component'
    dset.units = 'W.m-2.sr-1.um-1'
    dset = sgrp.createVariable('I_noise', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views', 'intensity_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of I'
    dset.units = 'W.m-2.sr-1.um-1'

    chunksizes = None if number_of_images is not None else \
        (256, n_bins_across, n_views, n_bins_polar)
    dset = sgrp.createVariable('I_polsample', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'I at polarimeter sampling'
    dset.units = 'W.m-2.sr-1.um-1'
    dset = sgrp.createVariable('I_polsample_noise', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of I at polarimeter sampling'
    dset.units = 'W.m-2.sr-1.um-1'
    dset = sgrp.createVariable('Q', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'Q Stokes component'
    dset.units = 'W.m-2.sr-1.um-1'
    dset = sgrp.createVariable('Q_noise', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of Q'
    dset.units = 'W.m-2.sr-1.um-1'
    dset = sgrp.createVariable('U', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'U Stokes component'
    dset.units = 'W.m-2.sr-1.um-1'
    dset = sgrp.createVariable('U_noise', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of U'
    dset.units = 'W.m-2.sr-1.um-1'
    dset = sgrp.createVariable('q', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'Q over I Stokes component'
    dset.units = '1'
    dset = sgrp.createVariable('q_noise', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of q'
    dset.units = '1'
    dset = sgrp.createVariable('u', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'U over I Stokes component'
    dset.units = '1'
    dset = sgrp.createVariable('u_noise', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of u'
    dset.units = '1'
    dset = sgrp.createVariable('DoLP', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'degree of linear polarization'
    dset.units = '1'
    dset = sgrp.createVariable('DoLP_noise', 'f4',
                               ('bins_along_track', 'bins_across_track',
                                'number_of_views',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'noise of DoLP'
    dset.units = '1'

    return rootgrp


# --------------------------------------------------
if __name__ == '__main__':
    fid = init_l1c('PACE_SPEX.20230115T123456.L1C.5km.V01.nc', {})
    fid.fill_global_attrs(orbit=12345, bin_size='5km')
    fid.close()



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Class to read SPEXone ICU packages

Copyright (c) 2019-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import numpy as np

from pyspex.lib.tmtc_def import tmtc_def

# - global parameters ------------------------------


# - local functions --------------------------------


# - class CCSDSio -------------------------
class CCSDSio:
    """
    Defines SPEXone L0 data package

    Doc: TMTC handbook (SPX1-TN-005), issue 12, 2020-05-15
    """
    def __init__(self, tmtc_issue=12, verbose=False):
        """
        Parameters
        ----------
        tmtc_issue: int
           Issue of the TMTC handbook which contains the definition of the
           Science Data Header format. Default: 12
        verbose: bool
           Be verbose. Default: be not verbose
        """
        # initialize class attributes
        self.offset = 0
        self.__hdr = None
        self.segmented = True
        self.tmtc_issue = tmtc_issue
        self.verbose = verbose

    @staticmethod
    def __hdr1_def():
        """
        Defines parameters of Primary header
        - Packet type     (3 bits): Version No.
                                    Indicates this is a CCSDS version 1 packet
                           (1 bit): Type indicator
                                    Indicates this is a telemetery packet
                           (1 bit): Secondary flag
                                    Indicate presence of Secondary header
                         (11 bits): ApID
                                    SPEXone ApID [0x320 - 0x351] or 2047

        - Packet Sequence (2 bits): Grouping flag
                                    00 continuation packet-data segment
                                    01 first packet-data segment
                                    10 last packet-data segment
                                    11 packet-data unsegmented
                         (14 bits): Counter per ApID, rollover to 0 at 0x3FFF
        - Packet length  (16 bits): size of packet data in bytes (always odd)
                                    (secondary header + User data) - 1
        """
        return [
            ('type', '>u2'),         # 0x000
            ('sequence', '>u2'),     # 0x002
            ('length', '>u2')        # 0x004
        ]

    @staticmethod
    def __timestamp():
        """
        Defines parameters of a timestamp
        - Seconds     (32 bits): seconds (TAI)
        - Sub-seconds (16 bits): sub-seconds (1/2 ** 16)
        """
        return [
            ('tai_sec', '>u4'),      # 0x006
            ('sub_sec', '>u2')       # 0x00A
        ]

    @property
    def version_no(self):
        """
        Returns CCSDS version number
        """
        if self.__hdr is None:
            return None

        return (self.__hdr['type'] >> 13) & 0x7

    @property
    def type_indicator(self):
        """
        Returns type of telemetry packet
        """
        if self.__hdr is None:
            return None

        return (self.__hdr['type'] >> 12) & 0x1

    @property
    def secnd_hdr_flag(self):
        """
        Returns flag indicating presence of a secondary header
        """
        if self.__hdr is None:
            return None

        return (self.__hdr['type'] >> 11) & 0x1

    @property
    def ap_id(self):
        """
        Returns SPEXone ApID
        """
        if self.__hdr is None:
            return None

        return self.__hdr['type'] & 0x7FF

    @property
    def grouping_flag(self):
        """
        Returns grouping flag

        Possible values:
          00 continuation packet-data segment
          01 first packet-data segment
          10 last packet-data segment
          11 packet-data unsegmented
        """
        if self.__hdr is None:
            return None

        return (self.__hdr['sequence'] >> 14) & 0x3

    @property
    def sequence_count(self):
        """
        Returns sequence counter, rollover to zero at 0x3FFF
        """
        if self.__hdr is None:
            return None

        return self.__hdr['sequence'] & 0x3FFF

    @property
    def packet_length(self):
        """
        Returns size of packet data in bytes
          Value equals secondary header + user data (always odd)
        """
        if self.__hdr is None:
            return None

        return self.__hdr['length']

    def __tm(self, num_data):
        """
        Return empty telemetry packet
        """
        return np.zeros(1, dtype=np.dtype([
            ('primary_header', np.dtype(self.__hdr1_def())),
            ('secondary_header', np.dtype(self.__timestamp())),
            ('mps', np.dtype(tmtc_def(0x350))),
            ('icu_time', np.dtype(self.__timestamp())),
            ('image_data', 'u2', (num_data,))]))

    def __rd_tm_packet(self, flname):
        """
        Read next telemetry packet

        Parameters
        ----------
        None

        Returns
        -------
        TM packet: primary & secondary header, MPS and image-data
        """
        hdr1_dtype = np.dtype(self.__hdr1_def())
        hdr2_dtype = np.dtype(self.__timestamp())
        mps_dtype = np.dtype(tmtc_def(0x350))

        # read parts of one telemetry packet data
        with open(flname, 'rb') as fp:
            hdr_one = np.fromfile(fp, dtype=hdr1_dtype,
                                  count=1, offset=self.offset)
            if hdr_one.size == 0:
                self.offset = 0
                return None

            self.__hdr = hdr_one[0]
            if self.verbose:
                print('ApID: ', self.ap_id,
                      self.secnd_hdr_flag, self.grouping_flag,
                      self.sequence_count, self.packet_length)
            if self.ap_id != 0x350:
                return None

            hdr_two = None
            mps = None
            icu_time = None
            num_bytes = self.packet_length + 1
            if self.secnd_hdr_flag == 1:
                hdr_two = np.fromfile(fp, dtype=hdr2_dtype, count=1)[0]
                num_bytes -= hdr2_dtype.itemsize

            # MPS is provided in first segement or unsegmented data packet
            mps = None
            if self.grouping_flag in (1, 3):
                mps = np.fromfile(fp, dtype=mps_dtype, count=1)[0]
                num_bytes -= mps_dtype.itemsize
                if self.tmtc_issue == 12:
                    icu_time = np.fromfile(fp, dtype=hdr2_dtype, count=1)[0]
                    num_bytes -= hdr2_dtype.itemsize

                # Correct 32-bit integers which originate from 24-bit
                # Necessary due to an allignment problem, in addition,
                # the 4 bytes of DET_CHENA also contain DET_ILVDS
                key_list = ['DET_EXPTIME', 'DET_EXPSTEP', 'DET_KP1',
                            'DET_KP2', 'DET_EXPTIME2', 'DET_EXPSTEP2',
                            'DET_CHENA']
                mps['DET_ILVDS'] = mps['DET_CHENA'] & 0xf
                for key in key_list:
                    mps[key] = mps[key] >> 8

            # remainder is image data
            data = np.fromfile(fp, dtype='>u2', count=num_bytes // 2)
            if self.verbose:
                msg = '{:d} {:d} {:3d} {:d} {:5d} {:d} {:d} {:d}  {:9d}'
                if mps is None:
                    print(msg.format(
                        self.secnd_hdr_flag, self.grouping_flag,
                        self.ap_id, hdr_two['tai_sec'], hdr_two['sub_sec'],
                        fp.tell() - self.offset, data.nbytes,
                        self.packet_length, fp.tell()))
                else:
                    msg += ' {:3d} {:9d}'
                    print(msg.format(
                        self.secnd_hdr_flag, self.grouping_flag,
                        self.ap_id, hdr_two['tai_sec'], hdr_two['sub_sec'],
                        fp.tell() - self.offset, data.nbytes,
                        self.packet_length, fp.tell(),
                        mps['MPS_ID'], mps['IMRLEN']))

        # combine parts to telemetry packet
        tm_packet = self.__tm(data.size)
        tm_packet[0]['primary_header'] = self.__hdr
        if hdr_two is not None:
            tm_packet[0]['secondary_header'] = hdr_two
        if mps is not None:
            tm_packet[0]['mps'] = mps
        if icu_time is not None:
            tm_packet[0]['icu_time'] = icu_time
        tm_packet[0]['image_data'] = data

        # move offset to next telemetry packet
        self.offset += self.__hdr.nbytes + self.packet_length + 1
        return tm_packet[0]

    def read(self, flname):
        """
        Read science or house-keeping telemetry packages

        Parameters
        ----------
        flname: str
           Name of the file with SPEXone ICU packages

        Returns
        -------
        tuple with data packages
        """
        # We should check that segmented packages consist of a sequence
        # with grouping flags {1, N * 0, 2}. I need to clean-up this code!
        packets = ()
        while True:
            try:
                buff = self.__rd_tm_packet(flname)
            except (IOError, EOFError) as msg:
                print(msg)
                return None

            # end of loop
            if buff is None:
                break

            # check if data is segmented
            if not packets:
                self.segmented = (self.grouping_flag != 3)

            if self.segmented == (self.grouping_flag == 3):
                print("FATAL: mixing segmented and unsegmented packages")

            packets += (buff,)

        return packets

    def group(self, packets):
        """
        Combine segmented data-packages

        Parameters
        ----------
        packets: tuple
           Tuple with science or house-keeping telemetry packages

        Returns
        -------
        tuple with unsegmented or house-keeping telemetry packages
        """
        if not self.segmented:
            return packets

        if ((packets[0]['primary_header']['sequence'] >> 14) & 0x0003) != 1:
            msg = '[WARNING]: removed {:d} segments of incomplete first image'
            ii = 0
            for buff in packets:
                flag = (buff['primary_header']['sequence'] >> 14) & 0x0003
                if flag == 1:
                    break
                ii += 1

            print(msg.format(ii))
            packets = packets[ii:]

        if ((packets[-1]['primary_header']['sequence'] >> 14) & 0x0003) != 2:
            msg = '[WARNING]: removed {:d} segments of incomplete last image'

            ii = len(packets)
            while ii > 0:
                ii -= 1
                flag = (packets[ii]['primary_header']['sequence'] >> 14) \
                       & 0x0003
                if flag == 2:
                    break

            print(msg.format(len(packets) - (ii + 1)))
            packets = packets[:ii+1]

        res = ()
        prev_grp_flag = 2
        for buff in packets:
            # get grouping flag of current package
            new_grp_flag = (buff['primary_header']['sequence'] >> 14) & 0x0003

            # handle segmented data
            if new_grp_flag == 1:
                # group_flag of previous package should be 2
                if prev_grp_flag != 2:
                    print('[WARNING]: first segment, but previous not closed?')

                data_buffer = buff['image_data']
                buff0 = self.__tm(buff['mps']['IMRLEN'] // 2)[0]
                buff0['primary_header'] = buff['primary_header']
                buff0['secondary_header'] = buff['secondary_header']
                buff0['mps'] = buff['mps']
                res += (buff0,)
            elif new_grp_flag in (0, 2):
                # group_flag of previous package should be 0 or 1
                if prev_grp_flag not in (0, 1):
                    print('[WARNING]: previous packet not closed?')

                data_buffer = np.concatenate((data_buffer, buff['image_data']))
                if new_grp_flag == 2:
                    res[-1]['image_data'] = data_buffer

            # keep current group flag for next read
            prev_grp_flag = new_grp_flag

        return res



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Create a valid PACE SPEXone Level-1A product to estimate the size of a
science orbit product

Copyright (c) 2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timezone

import numpy as np

from pyspex.lv1_io import L1Aio
from pyspex.lib.tmtc_def import tmtc_def

# - global parameters ------------------------------


# - local functions --------------------------------

# - main function ----------------------------------
def main():
    """
    Generate a SPEXone inflight Level-1A product
    """
    start_time = datetime(2023, 1, 15, 12, 34, 56, tzinfo=timezone.utc)
    midnight = start_time.replace(hour=0, minute=0, second=0, microsecond=0)
    duration = 2880  # seconds

    # define product: name, level, version, start of time-coverage
    prod_level = 'L1A'
    bin_size = '2.5km'
    prod_version = 'V1.0'
    l1a_flname = 'PACE_SPEX.{:s}.{:s}.{:s}.{:s}.nc'.format(
        start_time.strftime('%Y%m%dT%H%M%S'),
        prod_level, bin_size, prod_version)

    # define size of the dimensions
    dims = {
        'samples_per_image': 194029,
    }

    # define MPS data and image data
    mps_data = np.zeros(1, dtype=np.dtype(tmtc_def(0x350)))
    hk_data = np.zeros(1, dtype=np.dtype(tmtc_def(0x320)))
    image = np.arange(dims['samples_per_image'], dtype=np.uint16).reshape(1, -1)

    with L1Aio(l1a_flname, dims=dims, inflight=True) as l1a:
        secnds = (1 + np.arange(3 * duration, dtype=float)) / 3\
            + (start_time - midnight).total_seconds()
        l1a.fill_time(secnds, midnight)

        for ii in range(3 * duration):
            # Detector image data
            l1a.set_dset('/science_data/detector_images', image)

            # Detector telemetry
            # - alternative when real MPS are available - l1a.fill_mps(mps_data)
            l1a.set_dset('/science_data/detector_telemetry', mps_data)
            l1a.set_dset('/image_attributes/binning_table', [3])
            l1a.set_dset('/image_attributes/digital_offset', [0])
            l1a.set_dset('/image_attributes/nr_coadditions', [5])
            l1a.set_dset('/image_attributes/exposure_time', [0.056])
            l1a.set_dset('/image_attributes/image_ID', [ii])

        secnds = np.arange(duration, dtype=float) \
            + (start_time - midnight).total_seconds()
        l1a.set_dset('/engineering_data/HK_tlm_time', secnds)

        # Engineering data
        for _ in range(duration):
            l1a.set_dset('/engineering_data/HK_telemetry', hk_data)
            l1a.set_dset('/engineering_data/temp_detector', [293.])
            l1a.set_dset('/engineering_data/temp_optics', [294.])

        # Navigation data
        secnds = 2 * np.arange(duration // 2, dtype=float) \
            + (start_time - midnight).total_seconds()
        l1a.set_dset('/navigation_data/att_time', secnds)
        l1a.set_dset('/navigation_data/orb_time', secnds + .0132)

        for _ in range(max(1, duration // 2)):
            l1a.set_dset('/navigation_data/adstate', [255])
            l1a.set_dset('/navigation_data/att_quat',
                         [[0.0729222, -0.016155, 0.1108807, 0.9910231]])
            l1a.set_dset('/navigation_data/orb_pos',
                         [[404721.7, 1117799.9, -6963719.7]])
            l1a.set_dset('/navigation_data/orb_vel',
                         [[7295.4116, 1605.6194, 682.1598]])

        # Global attributes
        l1a.fill_global_attrs(orbit=12345, bin_size=bin_size)

    # with L1Aio(l1a_flname, append=True) as l1a:
    #    print(l1a.ref_date)


# --------------------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Convert SPEXone binning table in csv format to netCDF4

Copyright (c) 2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse
from datetime import datetime
from pathlib import Path

import numpy as np
from netCDF4 import Dataset

from pys5p.s5p_plot import S5Pplot
from pys5p.tol_colors import tol_cmap

from pyspex import version


# - global parameters ------------------------------


# - local functions --------------------------------


# - main function ----------------------------------
def main():
    """
    main program to generate a netCDF4 file with a SPEXone binning table
    """
    parser = argparse.ArgumentParser(
        description=('create SPEXone L1A product from DEM measurement(s)'))
    parser.add_argument('--figure', default=False, action='store_true',
                        help='generate (PDF) figure of binning table(s)')
    parser.add_argument('--verbose', default=False, action='store_true',
                        help='be verbose, default be silent')
    parser.add_argument('file_list', nargs='+',
                        help=("provide path to SPEXone binning table(s)"
                              " in csv-format"))
    args = parser.parse_args()
    if args.verbose:
        print(args)

    # define FillValue
    fill_value = 0xFFFFFFFF  # 0X7FFFFFFF

    # initialize netCDF file with binning tables
    date_created = datetime.utcnow().strftime("%Y%m%dT%H%M%S")
    tbl_name = 'SPX1_OCAL_L1A_TBL_{}_{}.nc'.format(date_created, '00001')
    fid = Dataset(tbl_name, 'w')
    fid.setncattr('title', 'SPEXone Level-1 binning-tables')
    fid.setncattr('processing_version', version.get())
    fid.setncattr('date_created',
                  datetime.utcnow().isoformat(timespec='seconds'))
    fid.setncattr('Convensions', 'CF-1.6')
    fid.setncattr('project', 'PACE Project')
    fid.setncattr('instrument', 'SPEXone')
    fid.setncattr('institution',
                  'SRON Netherlands Institute for Space Research')

    fid.createDimension('row', 1024)
    fid.createDimension('column', 1024)

    # ----- define tables with table_id equals one -----
    table_id = 1
    # write binning table to netCDF4 file
    gid = fid.createGroup('Table_{:02d}'.format(table_id))
    gid.enabled_lines = np.uint16(1024)
    gid.flex_binned_pixels = np.uint32(0)

    dset = gid.createVariable('binning_table', 'u4', ('row', 'column'),
                              fill_value=fill_value, chunksizes=(128, 128),
                              zlib=True, complevel=1, shuffle=True)
    dset.long_name = 'binning table'
    dset.valid_min = np.uint32(0)
    dset.valid_max = np.uint32(0xfffff)
    dset[:] = np.arange(1024 ** 2, dtype='u4').reshape(1024, 1024)

    dset = gid.createVariable('coadding_table', 'u1', ('row', 'column'),
                              fill_value=0xFF, chunksizes=(128, 128),
                              zlib=True, complevel=1, shuffle=True)
    dset.long_name = 'number of coadditions'
    dset.valid_min = np.uint8(0)
    dset.valid_max = np.uint8(0xff)
    dset[:] = np.ones((1024, 1024), dtype='u1')

    # add binning tables
    for flname in args.file_list:
        flpath = Path(flname)
        if not flpath.is_file():
            raise FileNotFoundError(flname)
        if flpath.suffix != '.csv':
            continue

        # read csv-file
        table = np.loadtxt(flname, delimiter=',', dtype=np.uint32)

        # determine unique addresses and binning-counts
        _, count = np.unique(table, return_counts=True)

        # no valid binning-counts larger than 6
        indx = np.where(count > 6)[0]
        for ii in indx:
            table[table == ii] = fill_value

        # convert csv binning table to binning table with FillValues
        frame = np.full(table.shape, 0, dtype='u1')
        frame[table != fill_value] = count[table[table != fill_value]]

        # write binning table to netCDF4 file
        table_id += 1
        gid = fid.createGroup('Table_{:02d}'.format(table_id))
        gid.origin = flpath.name
        gid.enabled_lines = np.uint16(np.sum(np.sum(frame, axis=1) > 0))
        gid.flex_binned_pixels = np.uint32(np.sum(frame > 0))

        dset = gid.createVariable('binning_table', 'u4', ('row', 'column'),
                                  fill_value=fill_value, chunksizes=(128, 128),
                                  zlib=True, complevel=1, shuffle=True)
        dset.long_name = 'binning table'
        dset.valid_min = np.uint32(0)
        dset.valid_max = np.uint32(np.max(table[table <= 0x7ffff]))
        dset[:] = table

        dset = gid.createVariable('coadding_table', 'u1', ('row', 'column'),
                                  fill_value=0xFF, chunksizes=(128, 128),
                                  zlib=True, complevel=1, shuffle=True)
        dset.long_name = 'number of coadditions'
        dset.valid_min = np.uint8(0)
        dset.valid_max = np.uint8(0xff)
        dset[:] = frame

        if args.figure:
            plot = S5Pplot(flpath.with_suffix('.pdf').name)
            plot.draw_signal(table, add_medians=False,
                             vrange=[0, np.max(table[table <= 0x7ffff])],
                             title=flpath.name, sub_title='Binning Table')
            plot.set_cmap(tol_cmap('rainbow_discrete', 5))
            plot.draw_signal(frame, add_medians=False,
                             title=flpath.name, sub_title='Number of Coaddings')
            plot.close()

    # read line-skip definitions
    for flname in args.file_list:
        flpath = Path(flname)
        if not flpath.is_file():
            raise FileNotFoundError(flname)
        if flpath.suffix != '.dat':
            continue

        # read line-skip definition
        data = np.loadtxt(flname, delimiter=',', dtype=np.uint8)
        line_indx = np.where(data == 1)[0]
        indx = np.arange(1024 * line_indx.size).reshape(line_indx.size, -1)

        table = np.full((1024, 1024), fill_value, dtype='u4')
        table[line_indx, :] = indx

        frame = np.zeros((1024, 1024), dtype='u1')
        frame[line_indx, :] = 1

        # write binning table to netCDF4 file
        table_id += 1
        gid = fid.createGroup('Table_{:02d}'.format(table_id))
        gid.origin = flpath.name
        gid.enabled_lines = np.uint16(line_indx.size)
        gid.flex_binned_pixels = np.uint32(0)

        dset = gid.createVariable('binning_table', 'u4', ('row', 'column'),
                                  fill_value=fill_value, chunksizes=(128, 128),
                                  zlib=True, complevel=1, shuffle=True)
        dset.long_name = 'binning table'
        dset.valid_min = np.uint32(0)
        dset.valid_max = np.uint32(np.max(table[table != fill_value]))
        dset[:] = table

        dset = gid.createVariable('coadding_table', 'u1', ('row', 'column'),
                                  fill_value=0xFF, chunksizes=(128, 128),
                                  zlib=True, complevel=1, shuffle=True)
        dset.long_name = 'number of coadditions'
        dset.valid_min = np.uint8(0)
        dset.valid_max = np.uint8(0xff)
        dset[:] = frame

        if args.figure:
            plot = S5Pplot(flpath.with_suffix('.pdf').name)
            plot.draw_signal(table, add_medians=False,
                             vrange=[0, np.max(table[table <= 0x7ffff])],
                             title=flpath.name, sub_title='Binning Table')
            plot.set_cmap(tol_cmap('rainbow_discrete', 5))
            plot.draw_signal(frame, add_medians=False,
                             title=flpath.name, sub_title='Number of Coaddings')
            plot.close()

    fid.close()


# --------------------------------------------------
if __name__ == '__main__':
    main()
